import pickle
import time
from multiprocessing import Pool
from pathlib import Path

import numpy as np
from bs_python_utils.bsutils import final_s, mkdir_if_needed, print_stars
from cupid_matching.choo_siow import entropy_choo_siow
from cupid_matching.matching_utils import Matching, VarianceMatching
from cupid_matching.min_distance import estimate_semilinear_mde
from cupid_matching.model_classes import ChooSiowPrimitives
from cupid_matching.poisson_glm import choo_siow_poisson_glm

from config import (
    degrees,
    do_create_samples,
    do_simuls,
    do_simuls_mde,
    do_simuls_poisson,
    model_name,
    n_sim,
    nb_cpus,
    plot_simuls,
    renormalize_true_coeffs,
    results_dir,
    sample_size,
    samples_dir,
    shrink_factor,
    shrunk_model,
    use_multiprocessing,
    zero_guard,
)
from create_two_samples import make_sample
from plots import plot_simulation_results
from read_data import prepare_data_cupid
from simulate import run_simul
from specification import DGPSimul, make_bases, renormalize_dgp


def name_and_pickle_primitives(
    mus: Matching, varmus: VarianceMatching, base_functions: np.ndarray
) -> tuple[str, Path]:
    """pickles the population matching, its variance, and the base functions,
    and creates a qualified name

    Args:
        mus: the population Matching
        varmus: the corresponding variances
        base_functions: the values of the base functions

    Returns:
        the qualified name and the directory name for the results
    """
    full_model_name = f"{model_name}_{sample_size}_z{zero_guard}"
    subdir = f"{sample_size}_z{zero_guard}"
    if shrunk_model:
        full_model_name = f"{full_model_name}_s{shrink_factor}"
        subdir = f"{subdir}_s{shrink_factor}"
    with open(
        samples_dir / f"{full_model_name}_mus.pkl",
        "wb",
    ) as f:
        pickle.dump(mus, f)
    with open(samples_dir / f"{full_model_name}_varmus.pkl", "wb") as f:
        pickle.dump(varmus, f)
    with open(samples_dir / f"{full_model_name}_base_functions.pkl", "wb") as f:
        pickle.dump(base_functions, f)

    results_subdir = mkdir_if_needed(results_dir / subdir)

    return full_model_name, results_subdir


def estimate_original(
    mus: Matching,
    base_functions: np.ndarray,
    base_names: list[str],
) -> DGPSimul:
    """estimates Choo and Siow on the original data

    Args:
        mus: the population Matching
        base_functions: the values of the base functions
        base_names: the names of the base functions

    Returns:
        the estimated DGP
    """
    if do_simuls_mde:
        mde_results = estimate_semilinear_mde(mus, base_functions, entropy_choo_siow)
        estim_Phi = mde_results.estimated_Phi
        estim_coeffs = mde_results.estimated_coefficients
        varcov_coeffs = mde_results.varcov_coefficients
        std_coeffs = mde_results.stderrs_coefficients
    else:
        poisson_results = choo_siow_poisson_glm(mus, base_functions)
        estim_Phi = poisson_results.estimated_Phi
        estim_coeffs = poisson_results.estimated_beta
        var_gamma = poisson_results.varcov_gamma
        n_bases = base_functions.shape[2]
        varcov_coeffs = var_gamma[-n_bases:, -n_bases:]
        std_coeffs = np.sqrt(np.diag(varcov_coeffs))

    dgp_estimates = DGPSimul(
        Phi=estim_Phi,
        base_funs=base_functions,
        coeffs=estim_coeffs,
        base_names=base_names,
        stderrs=std_coeffs,
        varcov=varcov_coeffs,
    )

    print_stars("Original estimates of the DGP")
    print(dgp_estimates)

    return dgp_estimates


if __name__ == "__main__":
    do_both = do_simuls_mde and do_simuls_poisson

    if do_create_samples:
        make_sample(sample_size)

    mus, varmus = prepare_data_cupid(samples_dir, sample_size)
    muxy, mux0, mu0y, nx, my = mus.unpack()
    n_households_obs = np.sum(nx) + np.sum(my) - np.sum(muxy)
    base_functions, base_names = make_bases(nx, my, degrees)
    n_bases = len(base_names)
    full_model_name, results_subdir = name_and_pickle_primitives(
        mus, varmus, base_functions
    )

    if do_simuls:
        simulation_dgp = estimate_original(mus, base_functions, base_names)

        print(simulation_dgp)

        if renormalize_true_coeffs:
            simulation_dgp = renormalize_dgp(simulation_dgp)
            print_stars("  we normalized the DGP coefficients to 1")

        if shrunk_model:
            print_stars(f"  we shrink the DGP by a factor {shrink_factor: .2f}")
            simulation_dgp.base_funs /= shrink_factor
            simulation_dgp.Phi /= shrink_factor

        dgp_Phi = simulation_dgp.Phi
        dgp_base_functions = simulation_dgp.base_funs

        # we use this surplus and the margins we got from the Cupid dataset
        choo_siow_model = ChooSiowPrimitives(dgp_Phi, nx, my)

        # if we shrunk the model we need to reestimate it
        if shrunk_model:
            dgp_mus = choo_siow_model.simulate(n_households_obs, seed=564432)
            simulation_dgp = estimate_original(dgp_mus, dgp_base_functions, base_names)
            # and we renormalize once more
            simulation_dgp = renormalize_dgp(simulation_dgp)
            print_stars("  we normalized the DGP coefficients to 1")
        else:
            dgp_mus = mus

        with open(samples_dir / f"{full_model_name}_dgp_Phi.pkl", "wb") as f:
            pickle.dump(dgp_Phi, f)

        # generate random seeds
        rng = np.random.default_rng(130962)
        seeds = rng.integers(100_000, size=n_sim)
        verbose = 2

        # run simulation
        list_args = [
            (
                i_sim,
                seeds[i_sim],
                choo_siow_model,
                n_households_obs,
                dgp_base_functions,
                zero_guard,
                do_simuls_mde,
                do_simuls_poisson,
                verbose,
            )
            for i_sim in range(n_sim)
        ]

        if use_multiprocessing:
            start_simuls = time.perf_counter()
            with Pool(nb_cpus) as pool:
                results = pool.starmap(run_simul, list_args)
            end_simuls = time.perf_counter()
        else:
            nb_cpus = 1
            start_simuls = time.perf_counter()
            results = [None] * n_sim
            for i_sim in range(n_sim):
                results[i_sim] = run_simul(*list_args[i_sim])
            end_simuls = time.perf_counter()

        # unpack simulation results
        dgp_coeffs = simulation_dgp.coeffs
        dgp_stderrs = simulation_dgp.stderrs
        dgp_varcov = simulation_dgp.varcov

        if do_both:
            estim_coeffs_mde = np.zeros((n_sim, n_bases))
            estim_coeffs_poisson = np.zeros((n_sim, n_bases))
            for i_sim in range(n_sim):
                estim_coeffs_mde[i_sim, :] = results[i_sim][0]
                estim_coeffs_poisson[i_sim, :] = results[i_sim][1]
            simul_results = {
                "Base names": base_names,
                "Base functions": dgp_base_functions,
                "DGP coeffs": dgp_coeffs,
                "MDE": estim_coeffs_mde,
                "Poisson": estim_coeffs_poisson,
            }
        elif do_simuls_mde:
            estim_coeffs_mde = np.zeros((n_sim, n_bases))
            for i_sim in range(n_sim):
                estim_coeffs_mde[i_sim, :] = results[i_sim][0]
            simul_results = {
                "Base names": base_names,
                "Base functions": dgp_base_functions,
                "DGP coeffs": dgp_coeffs,
                "MDE": estim_coeffs_mde,
            }
        elif do_simuls_poisson:
            estim_coeffs_poisson = np.zeros((n_sim, n_bases))
            for i_sim in range(n_sim):
                estim_coeffs_poisson[i_sim, :] = results[i_sim][0]
            simul_results = {
                "Base names": base_names,
                "Base functions": dgp_base_functions,
                "DGP coeffs": dgp_coeffs,
                "Poisson": estim_coeffs_poisson,
            }
        simul_results["DGP stderrs"] = dgp_stderrs
        simul_results["DGP varcov"] = dgp_varcov
        phi_nonparam = np.zeros((n_sim, nx.size, my.size))
        for isim in range(n_sim):
            phi_nonparam[isim, :, :] = results[isim][-1]
        simul_results["Phi non param"] = phi_nonparam

        # and save them
        with open(
            results_subdir / f"{full_model_name}.pkl",
            "wb",
        ) as f:
            pickle.dump(simul_results, f)

        print(
            f"\n\n**** {n_sim} simulations took"
            f" {end_simuls - start_simuls: >10.3f} seconds on {final_s(nb_cpus, 'CPU')}****"
        )

    if plot_simuls:
        plot_simulation_results(
            full_model_name,
            results_subdir,
            expected_curve=True,
        )
