import pickle
from pathlib import Path

import numpy as np
from bs_python_utils.bsnputils import TwoArrays, print_quantiles
from cupid_matching.matching_utils import Matching, VarianceMatching, compute_margins

from config import zero_guard


def read_margins(
    samples_dir: Path,  # the samples directory
    sample_size: str,  # large or small sample
) -> TwoArrays:
    """reads and returns the margins for men and for women

    Args:
        samples_dir: the samples directory
        sample_size: "large" or "small" sample

    Returns:
        the margins for men, the margins for women
    """
    nx = np.loadtxt(samples_dir / f"{sample_size}_nx.txt")
    my = np.loadtxt(samples_dir / f"{sample_size}_my.txt")
    return nx, my


def read_marriages(
    samples_dir: Path,  # the data directory
    sample_size: str,  # large or small sample
) -> np.ndarray:
    """reads and returns the marriages and the variances

    Args:
        samples_dir: the samples directory
        sample_size: "large" or "small" sample

    Returns:
        the marriages matrix
    """
    muxy = np.loadtxt(samples_dir / f"{sample_size}_muxy.txt")
    return muxy


def reshape_varcov(
    varmus: np.ndarray,
    mus: Matching,
    n_households: float,
) -> VarianceMatching:
    """splits the variance-covariance matrix
    and renormalizes for a requested total number of households

    Args:
        varmus:  muxy row major, then  mux0, then mu0y packed in both dimensions
        mus: the original Matching
        n_households:  the number of households we want

    Returns:
         the 6 constituent blocks of the normalized variance-covariance
    """
    muxy, mux0, mu0y, *_ = mus.unpack()
    ncat_men, ncat_women = muxy.shape
    n_prod_categories = ncat_men * ncat_women
    # first we reshape
    varmus_xyzt = varmus[:n_prod_categories, :n_prod_categories]
    varmus_xyz0 = varmus[
        :n_prod_categories, n_prod_categories : (n_prod_categories + ncat_men)
    ]
    varmus_xy0t = varmus[:n_prod_categories, (n_prod_categories + ncat_men) :]
    varmus_x0z0 = varmus[
        n_prod_categories : (n_prod_categories + ncat_men),
        n_prod_categories : (n_prod_categories + ncat_men),
    ]
    varmus_x00y = varmus[
        n_prod_categories : (n_prod_categories + ncat_men),
        (n_prod_categories + ncat_men) :,
    ]
    varmus_0y0t = varmus[
        (n_prod_categories + ncat_men) :, (n_prod_categories + ncat_men) :
    ]
    # then we rescale
    n_households_mus = np.sum(muxy) + np.sum(mux0) + np.sum(mu0y)
    rescale_factor = n_households / n_households_mus
    rescale_factor2 = rescale_factor * rescale_factor

    varcovs = VarianceMatching(
        var_xyzt=varmus_xyzt * rescale_factor2,
        var_xyz0=varmus_xyz0 * rescale_factor2,
        var_xy0t=varmus_xy0t * rescale_factor2,
        var_x0z0=varmus_x0z0 * rescale_factor2,
        var_x00y=varmus_x00y * rescale_factor2,
        var_0y0t=varmus_0y0t * rescale_factor2,
    )

    return varcovs


def rescale_mus(
    mus: Matching,  # muxy, mux0, mu0y
    n_households: float,  # the number of households we want
) -> Matching:  # the normalized Matching after rescaling
    """normalizes the marriages and margins to a requested total number of households

    Args:
        mus: muxy, mux0, mu0y, etc
        n_households: the number of households we want

    Returns:
        the normalized Matching
    """
    muxy, mux0, mu0y, nx, my = mus.unpack()
    n_households_mus = np.sum(muxy) + np.sum(mux0) + np.sum(mu0y)
    rescale_factor = n_households / n_households_mus
    muxy_norm = muxy * rescale_factor
    nx_norm = nx * rescale_factor
    my_norm = my * rescale_factor
    mus_norm = Matching(muxy_norm, nx_norm, my_norm)
    return mus_norm


def _get_zeros_mu(mu: np.ndarray, eps: float = 1e-9) -> tuple[bool, int, float]:
    """finds the zero cells in an array `mu`

    Args:
        mu: an array
        eps: how small 0 is. Defaults to 1e-9.

    Returns:
        whether there are any zero cells; the size of the original array;
        and the smallest positive cell.
    """
    mu_size = mu.size
    nonzero_mu = mu[mu > eps]
    min_nonzero = np.min(nonzero_mu)
    n_zeros_mu = mu_size - nonzero_mu.size
    mu_has_zeros = n_zeros_mu > 0
    return mu_has_zeros, mu_size, min_nonzero


def boost_zero_cells(
    mus: Matching,  # muxy, mux0, mu0y, n, m
    coeff: float = 1.0,  # default scale factor for delta
) -> Matching:  # the transformed muxy, mux0, mu0y, nx, my
    """if `coeff` is not 0, add small number `delta` to 0-cells to avoid numerical issues

    Args:
        mus: muxy, mux0, mu0y, etc
        coeff: the default scale factor for delta

    Returns:
         the transformed muxy, mux0, mu0y, etc
    """
    muxy, mux0, mu0y, *_ = mus.unpack()
    zeros_muxy, muxy_size, min_muxy = _get_zeros_mu(muxy)
    zeros_mux0, mux0_size, min_mux0 = _get_zeros_mu(mux0)
    zeros_mu0y, mu0y_size, min_mu0y = _get_zeros_mu(mu0y)
    some_zeros = zeros_muxy or zeros_mux0 or zeros_mu0y
    if not some_zeros or coeff == 0:
        return mus
    else:
        delta = min(min_muxy, min_mux0, min_mu0y) / coeff
        muxy_fixed = muxy.astype(float)
        mux0_fixed = mux0.astype(float)
        mu0y_fixed = mu0y.astype(float)
        n_cells = 0
        if zeros_muxy:
            muxy_fixed += delta
            n_cells += muxy_size
        if zeros_mux0:
            mux0_fixed += delta
            n_cells += mux0_size
        if zeros_mu0y:
            mu0y_fixed += delta
            n_cells += mu0y_size
        n_households = np.sum(muxy) + np.sum(mux0) + np.sum(mu0y)
        scale_factor = n_households / (n_households + delta * n_cells)
        muxy_fixed *= scale_factor
        mux0_fixed *= scale_factor
        mux0_fixed *= scale_factor
        nx_fixed, my_fixed = compute_margins(muxy_fixed, mux0_fixed, mu0y_fixed)
        mus_fixed = Matching(muxy_fixed, nx_fixed, my_fixed)
        return mus_fixed


def prepare_data_cupid(
    samples_dir: Path, sample_size: str
) -> tuple[Matching, VarianceMatching]:
    """
    Prepares the data for estimation and simulation.

    Args:
        samples_dir: the samples directory
        sample_size: "large" or "small" sample

    Returns:
        the Matching and its estimated Variance
    """
    nx, my = read_margins(samples_dir, sample_size)
    muxy = read_marriages(samples_dir, sample_size)
    n_types_men, n_types_women = muxy.shape
    print(
        f"""
    \nThe data has {n_types_men: d} types of men and {n_types_women: d} types of women.
    """
    )
    mus = Matching(muxy, nx, my)
    mus_non_zero = boost_zero_cells(mus, coeff=zero_guard)

    with open(samples_dir / f"{sample_size}_varmus.pkl", "rb") as f:
        varmus = pickle.load(f)
    return mus_non_zero, varmus
