""" runs one simulation of the estimates """

import numpy as np
from cupid_matching.min_distance import estimate_semilinear_mde
from bs_python_utils.bsutils import final_s
from bs_python_utils.bsnputils import TwoArrays, ThreeArrays
from cupid_matching.model_classes import ChooSiowPrimitives
from cupid_matching.poisson_glm import choo_siow_poisson_glm
from cupid_matching.choo_siow import entropy_choo_siow

from config import n_simulations_done
from read_data import boost_zero_cells


def run_simul(
    i_sim: int,
    seed: int,
    choo_siow_model: ChooSiowPrimitives,
    n_households_obs: int,
    base_functions: np.ndarray,
    zero_guard: int,
    do_simuls_mde: bool,
    do_simuls_poisson: bool,
    verbose: int = 0,
) -> TwoArrays | ThreeArrays:
    """runs one simulation

    Args:
        i_sim: the index of the simulation
        seed: the seed for its random draws
        choo_siow_model: the model we simulate from
        n_households_obs: the number of households
        base_functions:  the bases
        zero_guard: the divider
        do_simuls_mde: run the MDE simulation
        do_simuls_poisson:  run the Poisson simulation
        verbose:  how verbose: 1 print simulation number,
                                2 print steps

    Returns:
        the Results object(s) from the simulation
        and the nonparametric estimate of the joint surplus
    """
    do_both = do_simuls_mde and do_simuls_poisson
    mus_sim = choo_siow_model.simulate(n_households_obs, seed=seed)

    # for MDE we add a small number to zero cells
    mus_sim_non0 = boost_zero_cells(mus_sim, coeff=zero_guard)

    if verbose >= 1:
        print(f"Doing simul {i_sim}")
    muxy, mux0, mu0y, *_ = mus_sim_non0.unpack()

    # phi_non_param is the nonparametric Choo and Siow 2006 estimator
    phi_nonparam = 2.0 * np.log(muxy) - np.log(mu0y)
    phi_nonparam -= np.log(mux0).reshape((-1, 1))
    
    if do_both:
        if verbose >= 2:
            print(f"    Doing MDE {i_sim}")
        mde_results_sim = estimate_semilinear_mde(mus_sim_non0, base_functions, entropy_choo_siow)
        estim_coeffs_mde = mde_results_sim.estimated_coefficients
        if verbose >= 2:
            print(f"    Done MDE {i_sim}")
            print(f"    Doing Poisson {i_sim}")
        poisson_results_sim = choo_siow_poisson_glm(mus_sim, base_functions, verbose=0)
        estim_coeffs_poisson = poisson_results_sim.estimated_beta
        if verbose >= 2:
            print(f"    Done Poisson {i_sim}")
    elif do_simuls_mde:
        if verbose >= 2:
            print(f"    Doing MDE {i_sim}")
        mde_results_sim = estimate_semilinear_mde(mus_sim_non0, base_functions, entropy_choo_siow)
        estim_coeffs_mde = mde_results_sim.estimated_coefficients
        if verbose >= 2:
            print(f"    Done MDE {i_sim}")
    elif do_simuls_poisson:
        if verbose >= 2:
            print(f"    Doing Poisson {i_sim}")
        poisson_results_sim = choo_siow_poisson_glm(mus_sim, base_functions, verbose=0)
        estim_coeffs_poisson = poisson_results_sim.estimated_beta
        if verbose >= 2:
            print(f"    Done Poisson {i_sim}")
    if verbose >= 1:
        print(f"        Done simul {i_sim}")

    with n_simulations_done.get_lock():
        n_simulations_done.value += 1  # type: ignore
        print(f"       Did {final_s(n_simulations_done.value, 'simulation')}")  # type: ignore

    if do_both:
        return (estim_coeffs_mde, estim_coeffs_poisson, phi_nonparam)
    elif do_simuls_mde:
        return (estim_coeffs_mde, phi_nonparam)
    else:                                      # do_simuls_poisson only
        return (estim_coeffs_poisson, phi_nonparam)
