from dataclasses import dataclass

import numpy as np
from bs_python_utils.bsnputils import (
    check_matrix,
    check_square,
    check_tensor,
    check_vector,
    legendre_polynomials,
    nprepeat_col,
    nprepeat_row,
    quantile_transform,
)
from bs_python_utils.bsutils import bs_error_abort


@dataclass
class DGPSimul:
    """stores the information about the DGP used in the simulations

    `Phi` is an (X,Y)-matrix with the values of the joint surplus
    `base_funs` is an (X, Y, K)-array with the values of the bases
    `coeffs` is a K-vector with their coefficients

    if provided,
        `base_names` is a list of the K names of the bases
        `stderrs` is a K-vector with the stderrs of the coefficients
        `varcov` is a (K, K)-matrix with their variance
    """

    Phi: np.ndarray
    base_funs: np.ndarray
    coeffs: np.ndarray
    base_names: list[str] | None = None
    stderrs: np.ndarray | None = None
    varcov: np.ndarray | None = None

    def __str__(self):
        _, base_funs, base_names, coeffs, stderrs, _ = self.unpack()
        X, Y, K = base_funs.shape
        repr_str = (
            f"This DGP has {X} types of men, {Y} of women, and {K} base functions"
        )
        repr_str += "\n its coefficients are:"
        if base_names and np.any(stderrs):
            for k in range(K):
                repr_str += (
                    f"\n{base_names[k]: >20}: {coeffs[k]: >10.3f} "
                    f" ({stderrs[k]: >10.3f})"
                )
        elif base_names:
            for k in range(K):
                repr_str += f"\n{base_names[k]: >20}: {coeffs[k]: >10.3f}"
        elif np.any(stderrs):
            for k in range(K):
                repr_str += f"\n{coeffs[k]: >10.3f} ({stderrs[k]: >10.3f})"
        else:
            for k in range(K):
                repr_str += f"\n{coeffs[k]: >10.3f}"
        return repr_str + "\n\n"

    def __post_init__(self):
        X, Y = check_matrix(self.Phi)
        X2, Y2, K = check_tensor(self.base_funs, 3, fun_name=None)
        if X2 != X or Y2 != Y:
            bs_error_abort(
                f"Phi is a ({X}, {Y}) matrix but base_funs is ({X2}, {Y2},{K})."
            )
        K2 = check_vector(self.coeffs)
        if K2 != K:
            bs_error_abort(f"coeffs is a {K2}-vector but base_funs is ({X}, {Y}, {K}).")
        if self.stderrs is not None:
            K3 = check_vector(self.stderrs)
            if K3 != K:
                bs_error_abort(
                    f"stderrs is a {K3}-vector but base_funs is ({X}, {Y}, {K})."
                )
        if self.varcov is not None:
            K4 = check_square(self.varcov, fun_name=None)
            if K4 != K:
                bs_error_abort(
                    f"varcov is ({K4}, {K4}) but base_funs is ({X}, {Y}, {K})."
                )
        if self.base_names is not None:
            K5 = len(self.base_names)
            if K5 != K:
                bs_error_abort(
                    f"base_names has {K5} elements but base_funs is ({X}, {Y}, {K})."
                )

    def unpack(self):
        return (
            self.Phi,
            self.base_funs,
            self.base_names,
            self.coeffs,
            self.stderrs,
            self.varcov,
        )


def renormalize_dgp(dgps: DGPSimul) -> DGPSimul:
    """normalize the coefficients of the DGP to 1

    Args:
        dgps: the starting DGP

    Returns:
        a renormalized DGP
    """
    Phi, base_funs, base_names, coeffs, stderrs, varcov = dgps.unpack()
    K = base_funs.shape[2]
    base_funs_renormalized = np.zeros_like(base_funs)
    for k in range(K):
        base_funs_renormalized[:, :, k] = base_funs[:, :, k] * coeffs[k]
    if stderrs is not None:
        stderrs /= np.abs(coeffs)
        varcov /= np.outer(coeffs, coeffs)
    dgp_renormalized = DGPSimul(
        Phi=Phi,
        base_funs=base_funs_renormalized,
        coeffs=np.ones_like(coeffs),
        base_names=base_names,
        stderrs=stderrs,
        varcov=varcov,
    )
    return dgp_renormalized


def shrink_dgp(dgps: DGPSimul, shrink_factor: float) -> DGPSimul:
    """reduce the range of variation of Phi by a multiplicative factor

    Args:
        dgps: the starting DGP
        shrink_factor: by how much we divide the base functions

    Returns:
        a shrunk DGP
    """
    Phi, base_funs, base_names, coeffs, stderrs, varcov = dgps.unpack()
    dgp_shrunk = DGPSimul(
        Phi=Phi / shrink_factor,
        base_funs=base_funs / shrink_factor,
        coeffs=coeffs,
        base_names=base_names,
        stderrs=stderrs,
        varcov=varcov,
    )
    return dgp_shrunk


def monomial_name(var: str, deg: int) -> str:
    """create a string with a nice name for a univariate term"""
    if deg == 0:
        return "1"
    if deg == 1:
        return var
    # deg > 1:
    return f"{var}^{deg}"


def make_base_name(deg_x: int, deg_y: int) -> str:
    """create a string with a nice name for a bivariate term

    Args:
        deg_x: degree in x
        deg_y: degree in y

    Returns:
        a string that looks like x^a y^b
    """
    if deg_x == 0:
        return monomial_name("y", deg_y)
    if deg_y == 0:
        return monomial_name("x", deg_x)
    # both degrees are positive
    return monomial_name("x", deg_x) + " " + monomial_name("y", deg_y)


def generate_bases(
    nx: np.ndarray,
    my: np.ndarray,
    degrees: list[tuple[int, int]],
) -> tuple[np.ndarray, list[str]]:
    """generates the bases for a semilinear specification

    Args:
        nx: the numbers of men of each type
        my: the numbers of women of each type
        degrees: the list of degrees for polynomials in `x` and `y`

    Returns:
         the matrix of base functions and their names
    """
    n_bases = 3 + len(degrees)
    n_types_men, n_types_women = nx.size, my.size
    types_men, types_women = np.arange(n_types_men), np.arange(n_types_women)
    base_funs = np.zeros((n_types_men, n_types_women, n_bases))
    base_names = [""] * n_bases
    base_funs[:, :, 0] = 1.0
    base_names[0] = "1"
    for y in types_women:
        base_funs[:, y, 1] = np.where(types_men > y, 1.0, 0.0)
        base_funs[:, y, 2] = np.where(types_men > y, types_men - y, 0)
    base_funs[:, :, 2] /= (n_types_men + n_types_women) / 2
    base_names[1] = "1(x>y)"
    base_names[2] = "max(x-y,0)"
    # we quantile-transform nx and my
    q_nx = quantile_transform(nx)
    q_my = quantile_transform(my)
    # and we use the Legendre polynomials on [0,1]
    max_deg_x = max(degree[0] for degree in degrees)
    max_deg_y = max(degree[1] for degree in degrees)
    polys_x = legendre_polynomials(q_nx, max_deg_x, a=0)
    polys_y = legendre_polynomials(q_my, max_deg_y, a=0)
    i_base = 3
    for deg_x, deg_y in degrees:
        poly_x = polys_x[:, deg_x]
        poly_y = polys_y[:, deg_y]
        base_funs[:, :, i_base] = np.outer(poly_x, poly_y)
        base_names[i_base] = make_base_name(deg_x, deg_y)
        i_base += 1
    return base_funs, base_names


def make_bases(
    nx: np.ndarray, my: np.ndarray, degrees: list[tuple[int, int]]
) -> tuple[np.ndarray, list[str]]:
    """create base functions

    Args:
        nx: numbers of men of each type
        my: numbers of women of each type
        degrees: degrees of the bivariate polynomials

    Returns:
        the values of the base functions and their names
    """
    base_functions, base_names = generate_bases(nx, my, degrees)
    n_bases = base_functions.shape[-1]
    print(f"We created {n_bases} bases:")
    for i_base, base_name in enumerate(base_names):
        print(f"{i_base+1}: {base_name}")
    return base_functions, base_names
